#include <string.h>
#include <stdlib.h>

#include "gp2x_sdl.h"
#include "gp2x_joy.h"
#include "gp2x_fs_dlg.h"
#include "gp2x_dir.h"

static int selected;
static int dirty;
static const char* dlg_title;

static int top_pos;         /* 0 based index of top most shown item */
static int current_pos;     /* 0 based index of selected item */

static dir_list* curlist = 0;

static TTF_Font* font      = NULL;
static SDL_Surface* screen = NULL;

static int down_held       = 0;
static int up_held         = 0;
static int left_held       = 0;
static int right_held      = 0;

/* used to catch timer hitting right after pressing down and quickly back up but timer sneaking in that small interval */

#define DEBOUNCE_NUM         2
static int hold_debounce   = DEBOUNCE_NUM;

#define SCROLL_INTERVAL      10

#define FG_COLOR  ( 0xFFFFFF )
#define BG_COLOR  ( 0x000000 )

#define TIMER_INTERVAL 100

static void gp2x_redraw_file_selector( SDL_Surface* surface );
static int gp2x_file_selector_event  ( SDL_Event* event );

static void gp2x_fs_handle_up   ( int is_input_down );
static void gp2x_fs_handle_down ( int is_input_down );
static void gp2x_fs_handle_left ( int is_input_down );
static void gp2x_fs_handle_right ( int is_input_down );
static void gp2x_fs_handle_enter( void );
static void gp2x_fs_handle_other( void );

static Uint32 gp2x_timer_callback( Uint32 interval, void* data );

///////////////////////////////////

char*
gp2x_show_file_selector( SDL_Surface* surface, TTF_Font* ttf_font, const char* start_dir, const char* title )
{
    /* TODO: Add support for extension filters 'a78;bin;rom', etc */
    /*       Add support for showing cwd when traversing */

    SDL_Event event;
    SDL_TimerID timer;

    int ret        = 1;
    char* filename = NULL;

    dlg_title      = title;
    selected       = 0;
    dirty          = 0;

    screen = surface;
    font   = ttf_font;

    if ( !screen )
    {
		printf( "Could not put up file selection dialog, no surface\n" );
		return NULL;
    }

    if ( !font )
    {
		printf( "Could not load file selection dialog font\n" );
		return NULL;
    }

	curlist = dirlist_load( start_dir );

    if ( !curlist )
    {
		printf( "Could not load read directory %s\n", start_dir );
		return NULL;
    }

	gp2x_redraw_file_selector( screen );

    /* timer used for held down key/button */
    timer = SDL_AddTimer( TIMER_INTERVAL, gp2x_timer_callback, NULL ); 

    if ( !timer )
		printf( "Could not add timer: scrolling while holding joystick up/down will not work\n" );

    while ( !selected )
    {
        if ( SDL_PollEvent( &event ) == 0 ) 
        {
            if ( dirty )
            {
                gp2x_redraw_file_selector( screen );
                dirty = 0;
            }

            continue;
        }

        ret = gp2x_file_selector_event( &event );

        if ( ret == 0 )     /* user cancelled out */
        {
            printf( "User is quitting file selection dialog with no selection\n" );
            break;
        }
    }

    SDL_RemoveTimer( timer );

    if ( selected )
    {
        filename = dirlist_get_path( curlist, current_pos );
        return strdup( filename );
    }

    return NULL;
}

///////////////////////////////////

static void
gp2x_redraw_file_selector( SDL_Surface* screen )
{
	SDL_Rect rect;
	int y;
	int i;
	char temp[6];
	char temp2[256];

	rect.x = rect.y = 0;
	rect.w = screen->w;
	rect.h = screen->h;

	SDL_FillRect( screen, &rect, BG_COLOR );

	gp2x_render_text_center_x( screen, font, dlg_title, 0, FG_COLOR );

	y = TTF_FontHeight( font );

	gp2x_render_text_center_x( screen, font, "Press A to select, Start to quit", y, FG_COLOR );

	y += ( TTF_FontHeight( font ) * 1.5 );

    sprintf( temp2, "[%s]", curlist->path );

	gp2x_render_text_center_x( screen, font, temp2, y, FG_COLOR );

	y += ( TTF_FontHeight( font ) * 1.5 );

	for ( i = top_pos; i < curlist->elem_count; i++ )
	{
        int w, h;

		dir_ent* entry = curlist->elements + i;
		
		rect.x = 0;
		rect.y = y;
		rect.w = screen->w;
		rect.h = TTF_FontHeight( font );

		SDL_FillRect( screen, &rect, BG_COLOR );

		if ( i == current_pos )
#ifdef __GP2X_HOST__
            /* bug on debian */
			sprintf( temp, "->__" );
#else
			sprintf( temp, "->  " );
#endif

		else
#ifdef __GP2X_HOST__
			sprintf( temp, "____" );
#else
			sprintf( temp, "    " );
#endif

        if ( entry->type == 'd' )
            sprintf( temp2, "%s/", entry->filename );
        else
            sprintf( temp2, "%s", entry->filename );

        TTF_SizeText( font, temp, &w, &h );

		gp2x_render_text( screen, font, temp, 0, y, FG_COLOR );
		gp2x_render_text( screen, font, temp2, w, y, FG_COLOR );

		y += TTF_FontHeight( font );

		if ( ( y + TTF_FontHeight( font ) ) >= screen->h )
			break;
	}

    SDL_UpdateRect( screen, 0, 0, screen->w, screen->h );

    if ( screen->flags & SDL_DOUBLEBUF )
        SDL_Flip( screen );
}

///////////////////////////////////

static int
gp2x_file_selector_event( SDL_Event* event )
{
    int ret = 1;        /* ret here means whether to keep going or not to caller */

    switch ( event->type )
    {
        case SDL_JOYBUTTONDOWN:
        {
            /*
            printf( "Button pressed down was %d\n", event->jbutton.button );
            fflush( stdout );
            */

            if ( event->jbutton.button == GP2X_BUTTON_DOWN )
                gp2x_fs_handle_down( 1 );

            else if ( event->jbutton.button == GP2X_BUTTON_UP )
                gp2x_fs_handle_up( 1 );

            else if ( event->jbutton.button == GP2X_BUTTON_LEFT )
                gp2x_fs_handle_left( 1 );

            else if ( event->jbutton.button == GP2X_BUTTON_RIGHT )
                gp2x_fs_handle_right( 1 );

            else if ( event->jbutton.button == GP2X_BUTTON_A )
                gp2x_fs_handle_enter();

            else if ( event->jbutton.button == GP2X_BUTTON_START )
                ret = 0;

            else
                break;      /* NOP, but don't exit either */

            break;

#if 0
            else if ( event->jbutton.button == GP2X_BUTTON_DOWNRIGHT )
            {
                //printf( "DownRight button pressed down\n" );
                gp2x_fs_handle_down( 1 );
            }

            else if ( event->jbutton.button == GP2X_BUTTON_DOWNLEFT )
            {
                //printf( "DownLeft button pressed down\n" );
                gp2x_fs_handle_down( 1 );
            }

            else if ( event->jbutton.button == GP2X_BUTTON_UPRIGHT )
            {
                //printf( "UpRight button pressed down\n" );
                gp2x_fs_handle_up( 1 );
            }

            else if ( event->jbutton.button == GP2X_BUTTON_UPLEFT )
            {
                //printf( "UpLeft button pressed down\n" );
                gp2x_fs_handle_up( 1 );
            }
#endif

        }

        case SDL_JOYBUTTONUP:
        {
            /*
            printf( "Button pressed up was %d\n", event->jbutton.button );
            fflush( stdout );
            */

#if 0
            else if ( event->jbutton.button == GP2X_BUTTON_DOWNRIGHT )
            {
                gp2x_fs_handle_down( 0 );
            }

            else if ( event->jbutton.button == GP2X_BUTTON_DOWNLEFT )
            {
                gp2x_fs_handle_down( 0 );
            }

            else if ( event->jbutton.button == GP2X_BUTTON_UPRIGHT )
            {
                //printf( "UpRight button pressed up\n" );
                gp2x_fs_handle_up( 0 );
            }

            else if ( event->jbutton.button == GP2X_BUTTON_UPLEFT )
            {
                //printf( "UpLeft button pressed up\n" );
                gp2x_fs_handle_up( 0 );
            }
#endif
            if ( event->jbutton.button == GP2X_BUTTON_DOWN )
                gp2x_fs_handle_down( 0 );

            else if ( event->jbutton.button == GP2X_BUTTON_UP )
                gp2x_fs_handle_up( 0 );

            else if ( event->jbutton.button == GP2X_BUTTON_LEFT )
                gp2x_fs_handle_left( 0 );

            else if ( event->jbutton.button == GP2X_BUTTON_RIGHT )
                gp2x_fs_handle_right( 0 );

            else
                break;      /* NOP, but don't exit either */

            break;
        }


#ifdef __GP2X_HOST__

        case SDL_KEYDOWN:

            if ( event->key.keysym.sym == SDLK_DOWN )
                gp2x_fs_handle_down( 1 );

            else if ( event->key.keysym.sym == SDLK_UP )
                gp2x_fs_handle_up( 1 );

            else if ( event->key.keysym.sym == SDLK_LEFT )
                gp2x_fs_handle_left( 1 );

            else if ( event->key.keysym.sym == SDLK_RIGHT )
                gp2x_fs_handle_right( 1 );

            else if ( event->key.keysym.sym == SDLK_RETURN )
                gp2x_fs_handle_enter();

            else
                ret = 0;

            break;

        case SDL_KEYUP:

            if ( event->key.keysym.sym == SDLK_DOWN )
                gp2x_fs_handle_down( 0 );

            else if ( event->key.keysym.sym == SDLK_UP )
                gp2x_fs_handle_up( 0 );

            else if ( event->key.keysym.sym == SDLK_LEFT )
                gp2x_fs_handle_left( 0 );

            else if ( event->key.keysym.sym == SDLK_RIGHT )
                gp2x_fs_handle_right( 0 );

            else if ( event->key.keysym.sym == SDLK_RETURN )
                break;      /* NOP, but don't exit either */

            else
                ret = 0;

            break;
#endif

        default:
            /* user events generated by SDL timer will hit here */
            gp2x_fs_handle_other();     /* used for holding down button/key */

            break;
    }

    return ret;
}

///////////////////////////////////

static void
gp2x_fs_handle_up( int is_input_down )
{
    if ( is_input_down )
    {
        up_held = 1;

        if ( current_pos > 0 )
        {
            current_pos--;

            if ( current_pos < top_pos )
                top_pos--;

            dirty = 1;
        }
    }

    else
    {
        dirty         = 1;
        up_held       = 0;
        hold_debounce = DEBOUNCE_NUM;
    }
}

///////////////////////////////////

static void 
gp2x_fs_handle_down( int is_input_down )
{
    if ( is_input_down )
    {
        down_held = 1;

        if ( current_pos < ( curlist->elem_count - 1 ) )
        {
            current_pos++;

            if ( ( current_pos - top_pos ) * TTF_FontHeight( font ) > ( screen->h - ( 5.5 * TTF_FontHeight ( font ) ) ) )
                top_pos++;

            dirty = 1;
        }
    }

    else
    {
        dirty         = 1;
        down_held     = 0;
        hold_debounce = DEBOUNCE_NUM;
    }
}

///////////////////////////////////

static void
gp2x_fs_handle_left( int is_input_down )
{
    if ( is_input_down )
    {
        left_held = 1;

        if ( current_pos > 0 )
        {
            int delta = current_pos - SCROLL_INTERVAL;

            printf( "[left] Current position %d, delta %d top_pos %d\n", current_pos, delta, top_pos );

            if ( delta >= 0 )
            {
                current_pos -= SCROLL_INTERVAL;
                delta = SCROLL_INTERVAL;
                printf( "[left] delta >= 0, current_pos %d new delta %d\n", current_pos, delta );
            }

            else
            {
                delta = top_pos;
                current_pos = 0;
                printf( "[left] delta < 0, current_pos %d new delta %d\n", current_pos, delta );
            }

            if ( current_pos < top_pos )
            {
                printf( "[left] top_pos %d, subtracting %d\n", top_pos, delta );
                top_pos -= delta;
                printf( "[left] top_pos %d\n", top_pos );

                if ( top_pos < 0 )
                    top_pos = 0;
            }

            dirty = 1;
        }
    }

    else
    {
        dirty         = 1;
        left_held     = 0;
        hold_debounce = DEBOUNCE_NUM;
    }
}

///////////////////////////////////

static void 
gp2x_fs_handle_right( int is_input_down )
{
    if ( is_input_down )
    {
        right_held = 1;

        if ( current_pos < ( curlist->elem_count - 1 ) )
        {
            int delta = ( curlist->elem_count - 1 ) - current_pos;

            if ( delta < SCROLL_INTERVAL )
                current_pos += delta;

            else
            {
                current_pos += SCROLL_INTERVAL;
                delta = SCROLL_INTERVAL;
            }

            if ( ( current_pos - top_pos ) * TTF_FontHeight( font ) > ( screen->h - ( 5.5 * TTF_FontHeight ( font ) ) ) )
                top_pos += delta;

            dirty = 1;
        }
    }

    else
    {
        dirty         = 1;
        right_held    = 0;
        hold_debounce = DEBOUNCE_NUM;
    }
}

///////////////////////////////////

static void 
gp2x_fs_handle_enter( void )
{
	char* temp;
	dir_list* new_list;
	const char* filename;
	
	if ( curlist->elements[current_pos].type != 'd' )
	{
		selected = 1;
		return;
	}

	filename = curlist->elements[current_pos].filename;

	if ( strcmp( filename, ".." ) == 0 )
	{
		char *p;
		
		temp = malloc( strlen( curlist->path ) + 1 );
		strcpy( temp, curlist->path );

		p = strrchr( temp, '/' );

		if ( p == NULL )
		{
			free( temp );
			return;
		}

		if ( p == temp )
			p++;
		*p = 0;
	}

	else
	{
		temp = dirlist_get_path( curlist, current_pos );
	}

	new_list = dirlist_load( temp );
	free( temp );

	if ( new_list )
	{
		dirlist_destroy( curlist );

		curlist = new_list;
		current_pos    = 0;
		top_pos        = 0;
		dirty          = 1;
	}
}

///////////////////////////////////

static void
gp2x_fs_handle_other( void )
{
    if ( up_held )
    {
        if ( hold_debounce )
            hold_debounce--;
        else
            gp2x_fs_handle_up( 1 );
    }

    if ( down_held )
    {
        if ( hold_debounce )
            hold_debounce--;
        else
            gp2x_fs_handle_down( 1 );
    }

    if ( left_held )
    {
        if ( hold_debounce )
            hold_debounce--;
        else
            gp2x_fs_handle_left( 1 );
    }

    if ( right_held )
    {
        if ( hold_debounce )
            hold_debounce--;
        else
            gp2x_fs_handle_right( 1 );
    }
}

///////////////////////////////////

static Uint32 
gp2x_timer_callback( Uint32 interval, void* data )
{
    SDL_Event event;
    event.type = SDL_USEREVENT;

    SDL_PushEvent( &event );

    return interval;
}
