#include <string.h>

#include "gp2x_sdl.h"

///////////////////////////////////

SDL_Surface*
gp2x_init_sdl( void )
{
	SDL_Surface* surface = NULL;

    //  NOTE: timer doesn't work on HW SDL yet
    if ( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_JOYSTICK | SDL_INIT_TIMER ) < 0 ) 
    {
		printf( "gp2x_init_sdl(): Could not initialize SDL. Error: %s\n", SDL_GetError() );
		return surface;
    }

	SDL_ShowCursor( 0 );
	SDL_JoystickOpen( 0 );

#ifdef __GP2X_HOST__
	surface = SDL_SetVideoMode( 320, 240, 8, SDL_SWSURFACE | SDL_ANYFORMAT );
#else
	surface = SDL_SetVideoMode( 320, 240, 16, SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_ANYFORMAT );
#endif

	if ( !surface ) 
		printf( "gp2x_init_sdl(): Could not create window. Error: %s\n", SDL_GetError() );

    return surface;
}

///////////////////////////////////

void
gp2x_fini_sdl( void )
{
	SDL_Quit();
}

///////////////////////////////////

TTF_Font*
gp2x_load_ttf_font( const char* font_file, int font_size )
{
	TTF_Font* font;

    int ret = TTF_Init();

    if ( ret < 0 )
    {
        printf( "gp2x_load_ttf_font(): Couldn't init TTF subsystem\n" );
		return NULL;
    }
	
	font = TTF_OpenFont( font_file, font_size );

	if ( font == 0 )
    {
        printf( "gp2x_load_ttf_font(): Couldn't load font %s, size %d\n", font_file, font_size );
		return NULL;
    }

	return font;
}

///////////////////////////////////

void
gp2x_unload_ttf_font( TTF_Font* font )
{
    if ( font )
        TTF_CloseFont( font );

    TTF_Quit();
}

///////////////////////////////////

void
gp2x_render_text_center_x( SDL_Surface* screen, TTF_Font* font, const char* string, int y, int color )
{
    int x, w, h;

	if ( y > screen->h )
		return;
    
    TTF_SizeText( font, string, &w, &h );

    x = ( screen->w - w ) / 2;

    gp2x_render_text( screen, font, string, x, y, color );
}

///////////////////////////////////

void
gp2x_render_text( SDL_Surface* screen, TTF_Font* font, const char* string, 
                  int x, int y, int color )
{
	SDL_Rect sr, dr;
    SDL_Surface* font_sur;
	int n, i, max, w, h;
	int char_width;
	int char_height;
    int ret;
	
	if ( x > screen->w || y > screen->h )
		return;

    /*  assumption here is we're using a fixed width font */

    TTF_SizeText( font, string, &w, &h );
	n   = strlen( string );

	char_width  = w / n;
	char_height = TTF_FontHeight( font );
	
	max = ( screen->w - x ) / char_width;

	if ( n > max )
		n = max;

    font_sur = TTF_RenderText_Solid( font, string, gp2x_translate_color( color ) );

    if ( !font_sur )
        printf("gp2x_render_text(): Error rendering font: %s\n", TTF_GetError());

	else
	{
		dr.x = x;
		dr.y = y;
		dr.w = w;
		dr.h = h;
		sr   = dr;
		sr.y = 0;
		sr.x = 0;

		ret = SDL_BlitSurface( font_sur, &sr, screen, &dr );

		SDL_FreeSurface( font_sur );
	}
}

///////////////////////////////////

void
gp2x_outlined_rect( SDL_Surface* screen, int x, int y, int width, int height, int color )
{
    /* there doesn't seem to be a line or outlined rect func so I draw 4 rects of approp width/height */

    SDL_Rect top;
    SDL_Rect bottom;
    SDL_Rect left;
    SDL_Rect right;

    top.x = x;
    top.y = y;
    top.w = width;
    top.h = 1;

    SDL_FillRect( screen, &top, color );

    right.x = x + width;
    right.y = y;
    right.w = 1;
    right.h = height;

    SDL_FillRect( screen, &right, color );

    bottom.x = x;
    bottom.y = y + height;
    bottom.w = width;
    bottom.h = 1;

    SDL_FillRect( screen, &bottom, color );

    left.x = x;
    left.y = y;
    left.w = 1;
    left.h = height;

    SDL_FillRect( screen, &left, color );
}

///////////////////////////////////

SDL_Color 
gp2x_translate_color( int int_color )
{
    //  Change from an "int color" to an SDL_Color
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    SDL_Color color = { (int_color & 0x00ff0000) / 0x10000, ( int_color & 0x0000ff00 ) / 0x100,( int_color & 0x000000ff ), 0 };
#else
    SDL_Color color = { ( int_color & 0x000000ff ), ( int_color & 0x0000ff00 ) / 0x100, ( int_color & 0x00ff0000 ) / 0x10000, 0 };
#endif                                            
    return color;
}
