#include <string.h>
#include "gp2x_vkeyb.h"
#include "gp2x_sdl.h"
#include "gp2x_joy.h"
#include "keycodes.h"
#include "mesintrf.h"
#include "sysdep/sysdep_display_priv.h"

#define KEYPRESS_TICK_DEBOUNCE 100

int gp2x_display_vkeyb     = 0; /* global on whether to show it or not */
int gp2x_clean_up_keyboard = 0; /* global on whether to clean up keybd, positive val is # of times to clear out vkeyb rect */

static gp2x_vkeyb_entry last_key_hit;

static SDL_Surface* keybsurf = NULL;
static int cursor_x          = -1;
static int cursor_y          = -1;
static int cursor_width      = -1;
static int cursor_height     = -1;
static int last_pressed      = -1;
static int last_ticks        = -1;
static int caps_lock         =  0;
static int a_hit             =  0;
static int b_hit             =  0;
static int click_hit         =  0;

static void handle_key( gp2x_virtual_keyboard* keyb );

///////////////////////////////////

void 
gp2x_vkeyb_init( gp2x_virtual_keyboard* keyb )
{
    int w, h, keyb_w, keyb_h;

    TTF_SizeText( keyb->font, keyb->keyboard[0], &w, &h );

    keyb_w = w + 4;
    keyb_h = ( keyb->num_vkeyb_rows * TTF_FontHeight( keyb->font ) ) + 4;

	keybsurf = SDL_CreateRGBSurface( SDL_SWSURFACE, keyb_w, keyb_h, 16, 

#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		0x000000FF, 0x0000FF00, 0x00FF0000, 0 );
#else
		0xFF000000, 0x00FF0000, 0x0000FF00, 0 );
#endif

    SDL_FillRect( keybsurf, NULL, keyb->bg_color );
    SDL_SetAlpha( keybsurf, SDL_SRCALPHA, keyb->alpha );
}

///////////////////////////////////

void
gp2x_vkeyb_cleanup( void )
{
    if ( keybsurf )
        SDL_FreeSurface( keybsurf );

    keybsurf = NULL;
}

///////////////////////////////////

void 
gp2x_vkeyb_draw( SDL_Surface* screen, gp2x_virtual_keyboard* keyb )
{
	int i;
    int vkeyb_x = 2;
    int vkeyb_y = 2;
	SDL_Rect sr, dr;
    char** cur_keyboard;
    int w, h, keyb_w, keyb_h, char_w;

    TTF_SizeText( keyb->font, keyb->keyboard[0], &w, &h );

    keyb_w = w;
    keyb_h = h;

    if ( cursor_x == -1 )
        cursor_x = keyb->start_x;

    if ( cursor_y == -1 )
        cursor_y = keyb->start_y;

    if ( cursor_width == -1 )
        cursor_width = keyb->start_width;

    if ( cursor_height == -1 )
        cursor_height = keyb->start_height;

    if ( last_ticks == -1 )
        last_ticks = SDL_GetTicks();

    SDL_FillRect( keybsurf, NULL, keyb->bg_color );

    /* keyboard, all lines on keyboard are assumed to be the same length */

    cur_keyboard = keyb->keyboard;

    if ( caps_lock )
        cur_keyboard = keyb->shift_keyb;
        
	for ( i = 0; cur_keyboard[i]; i++ )
    {
        /* keys, line by line */
        gp2x_render_text( keybsurf, keyb->font, cur_keyboard[i], vkeyb_x, vkeyb_y + ( i * keyb_h ), keyb->fg_color );
    }

    char_w = w / strlen( keyb->keyboard[0] );

    /* key_behav status */

    for ( i = 0; i < keyb->num_vkeyb_entries; i++ )
    {
        if (    ( keyb->entries[i].key_behav == GP2X_VKEYB_TOGGLE_KEY_HL         && keyb->entries[i].last_state )
             || ( keyb->entries[i].key_behav == GP2X_VKEYB_NORMAL_KEY_HL_TOGGLE  && keyb->entries[i].last_state ) )
        {
            gp2x_outlined_rect( keybsurf, ( vkeyb_x + char_w * keyb->entries[i].x ) - 1, vkeyb_y + ( keyb_h * keyb->entries[i].y + 1 ), 
                      ( char_w * keyb->entries[i].width ), keyb_h - 2, keyb->toggle_color );
        }
    }

	/* cursor */

    gp2x_outlined_rect( keybsurf, ( vkeyb_x + char_w * cursor_x ) - 2, vkeyb_y + ( keyb_h * cursor_y ), 
                      char_w * cursor_width + 2, keyb_h, keyb->cursor_color );

    vkeyb_x = ( ( screen->w - keybsurf->w ) / 2 );
    vkeyb_y = screen->h - keybsurf->h;

	dr.x = vkeyb_x;
	dr.y = vkeyb_y;
	dr.w = keybsurf->w;
	dr.h = keybsurf->h;

	sr.x = 0;
	sr.y = 0;
	sr.w = keybsurf->w;
	sr.h = keybsurf->h;

    SDL_BlitSurface( keybsurf, &sr, screen, &dr );
}

///////////////////////////////////

void
gp2x_vkeyb_reset( void )
{
    cursor_x      = -1;
    cursor_y      = -1;
    cursor_width  = -1;
    cursor_height = -1;
    last_pressed  = -1;
    last_ticks    = -1;
    caps_lock     =  0;
    a_hit         =  0;
    b_hit         =  0;
    click_hit     =  0;

    keybsurf      = NULL;
}

///////////////////////////////////

void
gp2x_vkeyb_handle_input( int button_pressed, gp2x_virtual_keyboard* keyb )
{
    int i = 0, j = 0, x = 0, y = 0;

    int max_x     = strlen( keyb->keyboard[0] );
    int max_y     = keyb->num_vkeyb_rows;
    int cur_ticks = 0;

    last_pressed = button_pressed;

    cur_ticks = SDL_GetTicks();

    if ( cur_ticks - last_ticks < KEYPRESS_TICK_DEBOUNCE )
        return;

    last_ticks = cur_ticks;

    switch ( button_pressed )
    {
        case GP2X_BUTTON_DOWN:
        {
            int closest_x = 999999;        /* a # x will not be */
            int found     = -1;

            for ( i = 0; i < keyb->num_vkeyb_entries; i++ )
            {
                if ( keyb->entries[i].y == cursor_y + 1 )
                {
                    /* check for directly below */

                    for ( j = cursor_x; j <= ( cursor_x + cursor_width - 1 ); j++ )
                    {
                        if ( keyb->entries[i].x <= j && j <= ( keyb->entries[i].x + keyb->entries[i].width - 1 ) )
                        {
                            found = i;
                            break;
                        }
                    }

                    /* favor going to the down/right (diagonally) */

                    for ( j = cursor_x; j <= max_x; j++ )
                    {
                        if ( ( keyb->entries[i].x + keyb->entries[i].width -1 ) >= j && keyb->entries[i].x < closest_x )
                        {
                            closest_x = keyb->entries[i].x;
                            found     = i;
                        }
                    }
                }
            }

            if ( found >= 0 )
            {
                cursor_x      = keyb->entries[ found ].x;
                cursor_y      = keyb->entries[ found ].y;
                cursor_width  = keyb->entries[ found ].width;
                cursor_height = keyb->entries[ found ].height;
            }
        }

        break;

        case GP2X_BUTTON_UP:
        {
            int closest_x = -1;        /* a # x will not be */
            int found     = -1;

            for ( i = 0; i < keyb->num_vkeyb_entries; i++ )
            {
                if ( keyb->entries[i].y == cursor_y - 1 )
                {
                    /* check for directly above */

                    for ( j = cursor_x; j <= ( cursor_x + cursor_width - 1 ); j++ )
                    {
                        if ( keyb->entries[i].x <= j && j <= ( keyb->entries[i].x + keyb->entries[i].width - 1 ) )
                        {
                            found = i;
                            break;
                        }
                    }

                    for ( j = cursor_x; j >= 0; j-- )
                    {
                        /* favor going to the left/up (diagonally) */

                        if ( keyb->entries[i].x <= j && keyb->entries[i].x > closest_x )
                        {
                            closest_x = keyb->entries[i].x;
                            found     = i;
                        }
                    }
                }
            }

            if ( found >= 0 )
            {
                cursor_x      = keyb->entries[ found ].x;
                cursor_y      = keyb->entries[ found ].y;
                cursor_width  = keyb->entries[ found ].width;
                cursor_height = keyb->entries[ found ].height;
            }
        }

        break;

        case GP2X_BUTTON_LEFT:
        {
            int closest_x = -1;        /* a # x will not be */
            int found     = -1;

            for ( i; i < keyb->num_vkeyb_entries; i++ )
            {
                if ( keyb->entries[i].x < cursor_x && keyb->entries[i].y <= cursor_y && cursor_y <= ( keyb->entries[i].y + keyb->entries[i].height - 1 ) )
                {
                    if ( keyb->entries[i].x > closest_x )
                    {
                        found = i;
                        closest_x = keyb->entries[i].x;
                    }
                }
            }

            if ( found >= 0 )
            {
                cursor_x      = keyb->entries[ found ].x;
                cursor_y      = keyb->entries[ found ].y;
                cursor_width  = keyb->entries[ found ].width;
                cursor_height = keyb->entries[ found ].height;
            }
        }

        break;

        case GP2X_BUTTON_RIGHT:
        {
            int closest_x = 99999;        /* a # x will not be */
            int found     = -1;

            for ( i; i < keyb->num_vkeyb_entries; i++ )
            {
                if ( keyb->entries[i].x > cursor_x && keyb->entries[i].y <= cursor_y && cursor_y <= ( keyb->entries[i].y + keyb->entries[i].height - 1 ) )
                {
                    if ( keyb->entries[i].x < closest_x )
                    {
                        found = i;
                        closest_x = keyb->entries[i].x;
                    }
                }
            }

            if ( found >= 0 )
            {
                cursor_x      = keyb->entries[ found ].x;
                cursor_y      = keyb->entries[ found ].y;
                cursor_width  = keyb->entries[ found ].width;
                cursor_height = keyb->entries[ found ].height;
            }
        }

        break;

        case GP2X_BUTTON_A:

            a_hit = 1;

            handle_key( keyb );

            break;

        case GP2X_BUTTON_B:

            b_hit = 1;

            // handle_key( keyb );

            break;

        case GP2X_BUTTON_CLICK:

            click_hit = 1;
            break;

        default:
            ;
    }
}

///////////////////////////////////

void
gp2x_vkeyb_clear_display( SDL_Surface* screen )
{
    /* This routine is needed to clean up area outside of emulation
     * visible area
     */

    SDL_Rect s;
    int ret;

    s.x = 0;
    s.y = 0;
    s.w = screen->w;
    s.h = screen->h;

    ret = SDL_FillRect( screen, &s, 0x0 );
}

///////////////////////////////////

static void 
handle_key( gp2x_virtual_keyboard* keyb )
{
    int i;
    struct sysdep_display_keyboard_event kevent;

    for ( i = 0; i < keyb->num_vkeyb_entries; i++ )
    {
        if ( keyb->entries[i].x == cursor_x && keyb->entries[i].y == cursor_y )
        {
            /* Don't let the user quit out by hitting ESC while in GUI mode,
             * then can exit via the menu option
             */

            if ( mess_ui_active() && keyb->entries[i].keycode == KEY_ESC )
                return;

            if ( keyb->entries[i].key_behav == GP2X_VKEYB_TOGGLE_KEY_HL )
            {
                kevent.press = !keyb->entries[i].last_state;
                keyb->entries[i].last_state = !keyb->entries[i].last_state;
            }

            else if ( keyb->entries[i].key_behav == GP2X_VKEYB_NORMAL_KEY_HL_TOGGLE )
            {
                kevent.press = 1;
                keyb->entries[i].last_state = !keyb->entries[i].last_state;
            }

            else
                kevent.press = 1;

            /* TODO: the caps lock key just doesn't work in some of the drivers (ti99_4a) is it scan time?? */

            if ( keyb->entries[i].keycode == KEY_LSHIFT || keyb->entries[i].keycode == KEY_RSHIFT || keyb->entries[i].keycode == KEY_CAPSLOCK )
            {
                /* TODO: Somehow have to handle 2 shifts being on and one going off but keeping caps lock on */
                caps_lock = keyb->entries[i].last_state;
            }

            kevent.scancode = keyb->entries[i].keycode;
            sysdep_display_params.keyboard_handler( &kevent );

            last_key_hit = keyb->entries[i];
            break;
        }
    }
}

///////////////////////////////////

void gp2x_vkeyb_pump_event( void )
{
    /* This routine is necessary to give the input driver to register a transition in
     * keystrokes.  The joystick handler is polled and a check is done for a button
     * being pressed.  A better solution would be to send in button down and button up
     * events into here.
     */

    struct sysdep_display_keyboard_event kevent;

    if ( a_hit )
    {
        if ( last_key_hit.key_behav == GP2X_VKEYB_NORMAL_KEY )
        {
            kevent.press = 0;
            kevent.scancode = last_key_hit.keycode;
            sysdep_display_params.keyboard_handler( &kevent );
        }

        a_hit = 0;
    }

#if 0
    if ( b_hit )
    {
        /* TODO, implicit SHIFT held down */
    }
#endif
}
